/*
** ATOP - System & Process Monitor 
** 
** The program 'atop' offers the possibility to view the activity of
** the system on system-level as well as process-level.
** 
** This source-file contains functions to read all relevant system-level
** figures.
** ================================================================
** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
** E-mail:      gerlof@ATComputing.nl
** Date:        November 1996
** LINUX-port:  June 2000
** 
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License as published by the
** Free Software Foundation; either version 2, or (at your option) any
** later version.
** 
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
** See the GNU General Public License for more details.
**
** $Log: photosyst.c,v $
** Revision 1.8  2002/07/24 11:13:38  gerlof
** Changed to ease porting to other UNIX-platforms.
**
** Revision 1.7  2002/07/11 09:12:41  root
** Parsing of /proc/meminfo made 2.5 proof.
**
** Revision 1.6  2002/07/10 05:00:21  root
** Counters pin/pout renamed to swin/swout (Linux conventions).
**
** Revision 1.5  2002/07/08 09:31:11  gerlof
** *** empty log message ***
**
** Revision 1.4  2002/07/02 07:36:45  gerlof
** *** empty log message ***
**
** Revision 1.3  2002/07/02 07:08:36  gerlof
** Recognize more disk driver types via regular expressions
**
** Revision 1.2  2002/01/22 13:40:11  gerlof
** Support for number of cpu's.
**
** Revision 1.1  2001/10/02 10:43:31  gerlof
** Initial revision
**
*/

static char rcsid[] = "$Id: photosyst.c,v 1.8 2002/07/24 11:13:38 gerlof Exp $";

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <regex.h>

#include "atop.h"
#include "photosyst.h"

#define	MAXCNT	64

static int	isrealdisk(struct xdskstat);

	
void
photosyst(struct sstat *si)
{
	register int	i, nr, nlines;
	unsigned long	cnts[MAXCNT];
	FILE 		*fp;
	char		linebuf[1024], nam[64];
	static char	part_stats = 1; /* per-partition statistics ? */
	unsigned int	pgsz = getpagesz();

	memset(si, 0, sizeof(struct sstat));

	/*
	** gather various general statistics from the file /proc/stat and
	** store them in binary form
	*/
	if ( (fp = fopen("/proc/stat", "r")) != NULL)
	{
		while ( fgets(linebuf, sizeof(linebuf), fp) != NULL)
		{
			nr = sscanf(linebuf,
			 "%s  %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
			 "%lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
			 "%lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
			 "%lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
			 "%lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu ",
			  	nam,
			  	&cnts[0],  &cnts[1],  &cnts[2],  &cnts[3],
			  	&cnts[4],  &cnts[5],  &cnts[6],  &cnts[7],
			  	&cnts[8],  &cnts[9],  &cnts[10], &cnts[11],
			  	&cnts[12], &cnts[13], &cnts[14], &cnts[15],
			  	&cnts[16], &cnts[17], &cnts[18], &cnts[19],
			  	&cnts[20], &cnts[21], &cnts[22], &cnts[23],
			  	&cnts[24], &cnts[25], &cnts[26], &cnts[27],
			  	&cnts[28], &cnts[29], &cnts[30], &cnts[31],
			  	&cnts[32], &cnts[33], &cnts[34], &cnts[35],
			  	&cnts[36], &cnts[37], &cnts[38], &cnts[39],
			  	&cnts[40], &cnts[41], &cnts[42], &cnts[43],
			  	&cnts[44], &cnts[45], &cnts[46], &cnts[47],
			  	&cnts[48], &cnts[49], &cnts[50], &cnts[51],
			  	&cnts[52], &cnts[53], &cnts[54], &cnts[55],
			  	&cnts[56], &cnts[57], &cnts[58], &cnts[59],
			  	&cnts[60], &cnts[61], &cnts[62], &cnts[63]);

			if (nr < 2)		/* headerline ? --> skip */
				continue;

			if ( strcmp("cpu", nam) == EQ)
			{
				si->cpu.utime	= cnts[0];
				si->cpu.ntime	= cnts[1];
				si->cpu.stime	= cnts[2];
				si->cpu.itime	= cnts[3];
				continue;
			}

			if ( strncmp("cpu", nam, 3) == EQ)
			{
				si->cpu.numcpu++;
				continue;
			}

			if ( strcmp("ctxt", nam) == EQ)
			{
				si->cpu.csw	= cnts[0];
				continue;
			}

			if ( strcmp("intr", nam) == EQ)
			{
				si->cpu.devint	= cnts[0];
				continue;
			}

			if ( strcmp("swap", nam) == EQ)
			{
				si->mem.swins	= cnts[0];
				si->mem.swouts	= cnts[1];
				continue;
			}
		}

		fclose(fp);

		if (si->cpu.numcpu == 0)
			si->cpu.numcpu=1;
	}

	/*
	** gather memory-related statistics from the file /proc/meminfo and
	** store them in binary form
	**
	** in the file /proc/meminfo a 2.4 kernel starts with two lines
	** headed by the strings "Mem:" and "Swap:" containing all required
	** fields
        ** if these lines are present we try to skip parsing the rest
	** of the lines; if these lines are not present we should get the
	** required field from other lines
	*/
	if ( (fp = fopen("/proc/meminfo", "r")) != NULL)
	{
		int	nrfields = 7;	/* number of fields to be filled */

		while ( fgets(linebuf, sizeof(linebuf), fp) != NULL && 
								nrfields > 0)
		{
			nr = sscanf(linebuf,
				"%s %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu\n",
				nam,
				&cnts[0], &cnts[1], &cnts[2], &cnts[3],
				&cnts[4], &cnts[5], &cnts[6], &cnts[7],
				&cnts[8], &cnts[9]);

			if (nr < 2)		/* headerline ? --> skip */
				continue;

			if ( strcmp("Mem:", nam) == EQ)
			{
				si->mem.physmem	 	= cnts[0] / pgsz;
				si->mem.freemem		= cnts[2] / pgsz;
				si->mem.sharedmem	= cnts[3] / pgsz;
				si->mem.buffermem	= cnts[4] / pgsz;
				si->mem.cachemem	= cnts[5] / pgsz;
				nrfields -= 5;
			}
			else	if ( strcmp("Swap:", nam) == EQ)
				{
					si->mem.totswap  = cnts[0] / pgsz;
					si->mem.freeswap = cnts[2] / pgsz;
					nrfields -= 2;
				}
			else	if (strcmp("MemTotal:", nam) == EQ)
				{
					si->mem.physmem	  = cnts[0] *1024/pgsz;
					nrfields--;
				}
			else	if (strcmp("MemFree:", nam) == EQ)
				{
					si->mem.freemem	  = cnts[0] *1024/pgsz;
					nrfields--;
				}
			else	if (strcmp("MemShared:", nam) == EQ)
				{
					si->mem.sharedmem = cnts[0] *1024/pgsz;
					nrfields--;
				}
			else	if (strcmp("Buffers:", nam) == EQ)
				{
					si->mem.buffermem = cnts[0] *1024/pgsz;
					nrfields--;
				}
			else	if (strcmp("Cached:", nam) == EQ)
				{
					si->mem.cachemem  = cnts[0] *1024/pgsz;
				}
			else	if (strcmp("SwapCached:", nam) == EQ)
				{
					si->mem.cachemem  += cnts[0] *1024/pgsz;
					nrfields--;
				}
			else	if (strcmp("SwapTotal:", nam) == EQ)
				{
					si->mem.totswap	  = cnts[0] *1024/pgsz;
					nrfields--;
				}
			else	if (strcmp("SwapFree:", nam) == EQ)
				{
					si->mem.freeswap  = cnts[0] *1024/pgsz;
					nrfields--;
				}
		}

		fclose(fp);
	}

	if ( (fp = fopen("/proc/net/snmp", "r")) != NULL)
	{
		nlines = 3;

		while ( fgets(linebuf, sizeof(linebuf), fp) != NULL &&
		                                               nlines > 0)
		{
			nr = sscanf(linebuf,
			 "%s  %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
			 "%lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
			 "%lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
			 "%lu %lu\n",
				nam,
				&cnts[0],  &cnts[1],  &cnts[2],  &cnts[3],
				&cnts[4],  &cnts[5],  &cnts[6],  &cnts[7],
				&cnts[8],  &cnts[9], &cnts[10],  &cnts[11],
				&cnts[12], &cnts[13], &cnts[14], &cnts[15],
				&cnts[16], &cnts[17], &cnts[18], &cnts[19],
				&cnts[20], &cnts[21], &cnts[22], &cnts[23],
				&cnts[24], &cnts[25], &cnts[26], &cnts[27],
				&cnts[28], &cnts[29], &cnts[30], &cnts[31],
				&cnts[32], &cnts[33], &cnts[34], &cnts[35],
				&cnts[36], &cnts[37], &cnts[38], &cnts[39]);

			if (nr < 2)		/* headerline ? --> skip */
				continue;

			if ( strcmp("Ip:", nam) == EQ)
			{
				si->net.ipin	= cnts[2];
				si->net.ipout	= cnts[9];
				si->net.ipindel	= cnts[8];
				si->net.ipfrw	= cnts[5];
				nlines--;
				continue;
			}

			if ( strcmp("Tcp:", nam) == EQ)
			{
				si->net.tcpin	= cnts[9];
				si->net.tcpout	= cnts[10];
				nlines--;
				continue;
			}

			if ( strcmp("Udp:", nam) == EQ)
			{
				si->net.udpin	= cnts[0];
				si->net.udpout	= cnts[3];
				nlines--;
				continue;
			}
		}

		fclose(fp);
	}

	if ( (fp = fopen("/proc/net/dev", "r")) != NULL)
	{
		char *cp;

		i = 0;

		while ( fgets(linebuf, sizeof(linebuf), fp) != NULL)
		{
			if ( (cp = strchr(linebuf, ':')) != NULL)
				*cp = ' ';      /* substitute ':' by space */

			nr = sscanf(linebuf,
		 	    "%15s %lu %lu %*lu %*lu %*lu %*lu %*lu %*lu "
			    "%lu %lu %*lu %*lu %*lu %*lu %*lu %*lu",
				  si->intf[i].name,
				&(si->intf[i].rbyte),
				&(si->intf[i].rpack),
				&(si->intf[i].sbyte),
				&(si->intf[i].spack));

			if (nr == 5)	/* skip header & lines without stats */
			{
				if (++i >= MAXINTF-1)
					break;
			}
		}

		si->intf[i].name[0] = '\0';	/* set terminator for table */

		fclose(fp);
	}

	/*
	** check if extended partition-statistics are provided
	*/
	if ( part_stats && (fp = fopen("/proc/partitions", "r")) != NULL)
	{
		i = 0;

		while ( fgets(linebuf, sizeof(linebuf), fp) )
		{
			nr = sscanf(linebuf,
			      "%*d  %*d  %*lu %31s %lu %*lu %lu %*lu %lu "
			      "%*lu %lu %*lu %*lu  %lu %lu",
				 si->xdsk[i].name,
				&si->xdsk[i].nread,
				&si->xdsk[i].nrblk,
				&si->xdsk[i].nwrite,
				&si->xdsk[i].nwblk,
				&si->xdsk[i].io_ms,
				&si->xdsk[i].avque);

			/*
			** check if this line concerns the entire disk
			** or just one of the partitions of a disk (to be
			** skipped)
			*/
			if (nr == 7)	/* full stats-line ? */
			{
				if ( !isrealdisk(si->xdsk[i]) )
				       continue;
			
				if (++i >= MAXDSK-1)
					break;
			}
		}

		si->xdsk[i].name[0] = '\0';	/* set terminator for table */

		fclose(fp);

		if (i == 0)
			part_stats = 0;	/* do not try again for next cycles */
	}
}

static struct {
	char 	*regexp;
	regex_t	compreg;
} disknames[] = {
	"^sd[a-z][a-z]*$",			{0},
	"^hd[a-z]$",				{0},
	"^rd/c[0-9][0-9]*d[0-9][0-9]*$",	{0},
	"^cciss/c[0-9][0-9]*d[0-9][0-9]*$",	{0},
};

static int
isrealdisk(struct xdskstat ds)
{
	static int	firstcall = 1;
	register int	i;

	if (firstcall)		/* compile the regular expressions */
	{
		for (i=0; i < sizeof disknames/sizeof disknames[0]; i++)
			regcomp(&disknames[i].compreg, disknames[i].regexp,
								REG_NOSUB);

		firstcall = 0;
	}

	/*
	** try to recognize one of the compiled regular expressions
	*/
	for (i=0; i < sizeof disknames/sizeof disknames[0]; i++)
	{
		if (regexec(&disknames[i].compreg, ds.name, 0, NULL, 0) == 0)
			return 1;
	}

	return 0;
}
